# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

from LineGame import LineGameState
import GridUtils

# keys are positions to click, values are (lower left of quadrant position, direction (clockwise=1, ccw=-1)
ROTATION_POSITIONS = {(0,2): ((0,0), -1), (2,2): ((0,0), 1), (3,2): ((3,0), -1), (5,2): ((3,0), 1),
                      (0,5): ((0,3), -1), (2,5): ((0,3), 1), (3,5): ((3,3), -1), (5,5): ((3,3), 1),}
ROTATION_POSITION_ITEMS = list(ROTATION_POSITIONS.items())

def rotated_indexes(size, direction):
    """ Yields (from, to) tuples of positions indicating the source and destination indexes following a 
        90 degree rotation. direction=1 gives clockwise, direction=-1 gives counterclockwise. """
    for x in xrange(size):
        for y in xrange(size):
            rpos = (size-1-y, x) if direction==1 else (y, size-1-x)
            yield (rpos, (x,y))


def rotate_subgrid(grid, rdef, size):
    bottomleft, direction = rdef
    subgrid = grid[bottomleft[0]:bottomleft[0]+size, bottomleft[1]:bottomleft[1]+size].copy()
    for fromindex, toindex in rotated_indexes(size, direction):
        grid[GridUtils.add_tuples(bottomleft, toindex)] = subgrid[fromindex]
    return grid

class PentagoGameState(LineGameState):

    def win_length(self):
        return 5
        
    def subgrid_size(self):
        return self.grid.shape[0]//2;
        
    def successors(self):
        size = self.subgrid_size()
        for pos in self.enumerate_coords_with_value(0):
            # do each of 8 possible rotations (future optimization for AI: ignore effective duplicates)
            for rpos, rdef in ROTATION_POSITION_ITEMS:
                nextgrid = self.grid.copy()
                nextgrid[pos] = self.current_player
                rotate_subgrid(nextgrid, rdef, size)
                yield ( (pos, rpos), self.next_state_with_grid(nextgrid) )
                
    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        animations = []
        # create a new segment for the drop, we may need to add a rotation segment
        dropsegments = [PGPieceAnimation.AnimationSegment(action=PGPieceAnimation.NEW, from_position=move[0], from_value=self.current_player, start_time=0, end_time=PGPieceAnimation.DEFAULT_DURATION)]
        
        bottomleft, direction = ROTATION_POSITIONS[move[1]]
        # add animations only for rotating cells that aren't empty (avoids pause if entire subgrid is empty)
        for fromindex, toindex in rotated_indexes(self.subgrid_size(), direction):
            if fromindex==toindex: continue
            frompos = GridUtils.add_tuples(bottomleft, fromindex)
            oldval = self.grid[frompos]
            topos = GridUtils.add_tuples(bottomleft, toindex)
            if oldval:
                animations.append(PGPieceAnimation.move_animation(frompos, topos, dropsegments[0].end_time))
            elif frompos==move[0]:
                # add animation move segment for the new piece
                dropsegments.append(PGPieceAnimation.AnimationSegment(action=PGPieceAnimation.MOVE, from_value=self.current_player, from_position=move[0], to_position=topos, start_time=dropsegments[0].end_time, end_time=dropsegments[0].end_time+PGPieceAnimation.DEFAULT_DURATION))
        
        animations.append(PGPieceAnimation.Animation(None, dropsegments))
        return PGPieceAnimation.AnimationGroup(animations)
        
    
def test():
    import numpy
    grid = numpy.array(range(36)).reshape((6,6))
    subgrid = grid[0:3,0:3]
    assert list(subgrid.flat)==[0,1,2,6,7,8,12,13,14]
    rotate_subgrid(grid, ((0,0), 1), 3)
    print grid    
    rotate_subgrid(grid, ((0,0), -1), 3)
    print grid    
    rotate_subgrid(grid, ((0,0), -1), 3)
    print grid
    